#!/bin/bash

mkfifo --mode=700 /tmp/pipe-in$$ /tmp/pipe-out$$
exec 3<> /tmp/pipe-in$$
exec 4<> /tmp/pipe-out$$
rm /tmp/pipe-in$$ /tmp/pipe-out$$

{ netcat localhost 2010 ; echo 499 disconnected ; } <&3 >&4 &
pid=$!


function error() {
    SendCmd Quit

    exec 3>&-
    exec 4>&-

    kill $pid

    exit 1
}

function ReadReply() {
    reply2xx=()
    reply3xx=()
    reply4xx=()

    while read -r code line <&4 ; do
        echo "< $code $line"
        case $code in
            2*)     IFS=$' \t\n\r' reply2xx=($code "$line")
                    ;;
            3*)     IFS=$' \t\n\r' reply3xx=($code $line)
                    ;;
            4*)     IFS=$' \t\n\r' reply4xx=($code "$line")
                    ;;
        esac
        [ -n "${reply2xx[0]}" ] && break;
    done

    [ -n "${reply4xx[0]}" ] && return 1
    return 0
}

function SendCmd() {
    echo "> $*"
    echo "$*" >&3

    ReadReply
}


function HelloWorld() {
    SendCmd 'menu=NewMenu "Hello world, what do you think?"' || return 1
    SendCmd 'opt1=menu.AddOsdItem "This is awesome!"' || return 1
    SendCmd "opt2=menu.AddOsdItem \"You ain't seen nothing yet\"" || return 1
    SendCmd 'opt3=menu.AddOsdItem "All your OSD are belong to us"' || return 1
    SendCmd 'menu.Show' || return 1
    SendCmd 'menu.EnableEvent keyOk close' || return 1
    SendCmd 'menu.SleepEvent' || return 1

    if [ "${reply3xx[0]}" != "300" ] ; then return 1 ; fi

    if [ "${reply3xx[2]}" == "close" ] ; then
        SendCmd 'Message "You did not pick anything."' || return 1
        return 0
    fi

    if [ "${reply3xx[2]}" != "keyOk" ] ; then return 1 ; fi

    SendCmd 'menu.GetCurrent' || return 1
    if [ "${reply3xx[0]}" != "302" ] ; then return 1 ; fi

    case ${reply3xx[2]} in
        opt1)
            SendCmd 'Message "It is!"' || return 1
            ;;
        opt2)
            SendCmd 'Message "B-B-B-Baybee!"' || return 1
            ;;
        opt3)
            SendCmd 'Message "You are on the way to display!"' || return 1
            ;;
    esac
    SendCmd 'menu.SendState osEnd' || return 1
    return 0
}


ReadReply
SendCmd 'Version 0' || error

HelloWorld || error

SendCmd Quit

exec 3>&-
exec 4>&-

