

#include <vdr/config.h>
#include <vdr/skins.h>

#include "interpreter.h"


// ---------------
//   cConnection
// ---------------


cConnection::cConnection() {
}


char* cConnection::KeyString(char *buffer, int bufferlen, int key) {
    const char *k;

    if (key==-1) {
        strn0cpy(buffer,"timeout",bufferlen);
        return buffer;
    }

    if ((key&~k_Flags)==kNone) {
        k="None";
    } else {
        k=cKey::ToString((eKeys)(key & ~k_Flags));
        if (!k) k="Unknown";
    }
    switch (key&k_Flags) {
    case 0:
        snprintf(buffer,bufferlen,"key%s",k);
        break;
    case k_Repeat:
        snprintf(buffer,bufferlen,"key%s|Repeat",k);
        break;
    case k_Release:
        snprintf(buffer,bufferlen,"key%s|Release",k);
        break;
    case k_Repeat|k_Release:
        snprintf(buffer,bufferlen,"key%s|Repeat|Release",k);
        break;
    }
    return buffer;
}





// -------------------------------
//   cCommandProcessor::cCommand
// -------------------------------

cCommandProcessor::cCommand::cCommand() {
    Command=NULL;
    Context=NULL;
    AssignVar=NULL;
    argc=0;
}

bool cCommandProcessor::cCommand::ParseLine(cConnection &Connection, char *line) {
    char *p,*ps;

    // Reset parameters
    Command=NULL;
    Context=NULL;
    AssignVar=NULL;
    argc=0;
    argflagc=0;

    // Parse one identifier
    line=skipspace(line);
    if (*line==0 || *line=='#') return false;

    p=skipspace(ps=skipword(line));
    if (!p) {
        Connection.ReplyErrorIdentifier(line);
        return false;
    }

    if (*p=='=') {
        // Assignment found
        *ps=0;
        AssignVar=line;
        if (*line=='_') {
            Connection.ReplyErrorIdentifierReserved(line);
            return false;
        }
        // parse next word
        line=skipspace(p+1);
        p=skipspace(ps=skipword(line));
        if (!p) {
            Connection.ReplyErrorIdentifier(line);
            return false;
        }
    }

    if (*p=='.') {
        // Found member operator, parse member sequence and command
        char *lastdot;

        while (*p=='.') {
            // Loop a sequence of identifer.identifier.identifier.command
            lastdot=p;
            p=skipspace(ps=skipword(skipspace(p+1)));
            if (!p) {
                Connection.ReplyErrorIdentifier(line);
                return false;
            }
        }
        // Remember whole sequence up to before last dot
        *lastdot=0;
        Context=line;
        despace(Context);
        // Parse next word
        line=skipspace(lastdot+1);
        p=skipspace(ps=skipword(line));
        if (!p) {
            Connection.ReplyErrorIdentifier(line);
            return false;
        }
    }

    // Now parsed word is a command

    if (!isspace(*ps) && *ps!=0) {
        Connection.ReplyErrorIdentifier(line);
        return false;
    }
    *ps=0;

    Command=line;
    bool lastflag=false;

    while (*p!=0 && argc<MAXCOMMANDARGS) {
        char *q=p;
        bool quote=false;

        line=p;

        while (*p!=0 && (!isspace(*p) || quote)) {
            switch (*p) {
            case '\\':
                p++;
                if (*p=='n') *q++='\n';
                if (*p=='"') *q++='"';
                if (*p=='\\') *q++='\\';
                break;
            case '"':
                quote=!quote;
                break;
            default:
                *q++=*p;
            }
            p++;
        }
        p=skipspace(p);
        *q=0;
        if (!lastflag && 0==strcmp(line,"--")) {
            // Marker for end-of-flags reached
            lastflag=true;
            // ignore this parameter and dont increase argflagc from now on
            continue;
        }

        args[argc]=line;
        argc++;
        if (!lastflag) argflagc++;
    }

    return true;
}

bool cCommandProcessor::cCommand::ArgScanFlag(const char *flag) {
    int pos=0;

    while (pos<argflagc) {
        if (ArgIsFlag(pos) && 0==strcasecmp(flag,args[pos]+1)) break;
        pos++;
    }
    if (pos>=argflagc) return false;

    argc--;
    argflagc--;
    while (pos<argc) {
        args[pos]=args[pos+1];
        pos++;
    }
    return true;
}

char* cCommandProcessor::cCommand::ArgScanFlagParam(const char *flag) {
    int pos=0;

    while (pos<argflagc) {
        if (ArgIsFlag(pos) && 0==strcasecmp(flag,args[pos]+1)) break;
        pos++;
    }
    if (pos>=argflagc) return NULL;

    char *result=(pos+1<argc) ? (args[pos+1]) : (NULL);

    argc-=2;
    argflagc-=2;
    while (pos<argc) {
        args[pos]=args[pos+2];
        pos++;
    }
    return result;
}




// ---------------------
//   cCommandProcessor
// ---------------------


cCommandProcessor::cCommandProcessor(cConnection *connection) {
    Connection=connection;
    State=stateNew;
    SleepEventObj=NULL;
    SleepEventTimeout.tv_sec=0;
    SleepEventTimeout.tv_usec=0;
}

cCommandProcessor::~cCommandProcessor() {
    if (Connection) delete Connection;
}

bool cCommandProcessor::CmdVersion(cCommand &cmd) {
    if (cmd.GetContext()) {
        Connection->ReplyErrorUnknownCommand(cmd.GetContext());
        return false;
    }
    if (cmd.IsAssignment()) {
        Connection->ReplyErrorAssign();
        return false;
    }
    if (!cmd.HasMaxArgument(1)) {
        Connection->ReplyErrorUnknownParameter(cmd.GetArg(1));
        return false;
    }
    if (!cmd.HasMinArgument(1)) {
        Connection->ReplyErrorMissingParameter();
        return false;
    }

    if (0!=strcmp(cmd.GetArg(0),"0")) {
        Connection->ReplyErrorUnknownVersion(cmd.GetArg(0));
        return false;
    }
    return true;
}


bool cCommandProcessor::CmdQuit(cCommand &cmd) {
    if (cmd.IsAssignment()) {
        Connection->ReplyErrorAssign();
        return false;
    }
    if (!cmd.HasMaxArgument(0)) {
        Connection->ReplyErrorUnknownParameter(cmd.GetArg(0));
        return false;
    }
    Connection->ReplyGoodBye();
    return true;
}

bool cCommandProcessor::CmdIdle(cCommand &cmd) {
    if (cmd.IsAssignment()) {
        Connection->ReplyErrorAssign();
        return false;
    }
    if (!cmd.HasMaxArgument(0)) {
        Connection->ReplyErrorUnknownParameter(cmd.GetArg(0));
        return false;
    }
    return true;
}

bool cCommandProcessor::CmdMessage(cCommand &cmd) {
    /*
    MESSAGE [-info|-warn|-error] -queue [-timeout #] [-seconds #] message
    MESSAGE [-info|-warn|-error] [-seconds #] message
    MESSAGE -status message
    MESSAGE -statusclear
    */

    if (cmd.IsAssignment()) {
        Connection->ReplyErrorAssign();
        return false;
    }

    enum eMode { status, info, warn, error, statusclear }
        mode=info;

    int seconds=0;
    int timeout=0;

    bool queue=cmd.ArgScanFlag("queue");

    char *cseconds=cmd.ArgScanFlagParam("seconds");
    if (cseconds) seconds=atoi(cseconds);

    char *ctimeout=cmd.ArgScanFlagParam("timeout");
    if (ctimeout) timeout=atoi(ctimeout);

    if (ctimeout && !queue) {
        Connection->ReplyErrorParameter("-timeout requires -queue");
        return false;
    }


    int ArgCount=1;

    if (cmd.ArgScanFlag("status")) {
        mode=status;
        if (cseconds || queue) {
            Connection->ReplyErrorParameter("-status doesnt accept -queue or -seconds");
            return false;
        }
    } else if (cmd.ArgScanFlag("info")) {
        mode=info;
    } else if (cmd.ArgScanFlag("warn")) {
        mode=warn;
    } else if (cmd.ArgScanFlag("error")) {
        mode=error;
    } else if (cmd.ArgScanFlag("statusclear")) {
        mode=statusclear;
        if (cseconds || queue) {
            Connection->ReplyErrorParameter("-statusclear doesnt accept -queue or -seconds");
            return false;
        }
        ArgCount=0;
    }

    if (!cmd.HasMaxArgument(ArgCount)) {
        Connection->ReplyErrorUnknownParameter(cmd.GetArg(ArgCount));
        return false;
    }
    if (!cmd.HasMinArgument(ArgCount)) {
        Connection->ReplyErrorMissingParameter();
        return false;
    }

    if (queue) {
        int k=kNone;
        switch (mode) {
        case info:
            k=Skins.QueueMessage(mtInfo,cmd.GetArg(0),seconds,timeout);
            break;
        case warn:
            k=Skins.QueueMessage(mtWarning,cmd.GetArg(0),seconds,timeout);
            break;
        case error:
            k=Skins.QueueMessage(mtError,cmd.GetArg(0),seconds,timeout);
            break;
        default:;
        }

        if (timeout!=0) {
            Connection->ReplyMessageKeyEvent("Message",k);
        } else {
            Connection->ReplyMessageEventTimeout();
        }
    } else {
        int k=kNone;
        cMainThreadLock Lock;

        Lock.LockMainThread();
        switch (mode) {
        case status:
            k=Skins.Message(mtStatus,cmd.GetArg(0),0);
            break;
        case statusclear:
            k=Skins.Message(mtStatus,NULL,0);
            break;
        case info:
            k=Skins.Message(mtInfo,cmd.GetArg(0),seconds);
            break;
        case warn:
            k=Skins.Message(mtWarning,cmd.GetArg(0),seconds);
            break;
        case error:
            k=Skins.Message(mtError,cmd.GetArg(0),seconds);
            break;
        }
        Lock.UnlockMainThread();

        if (mode!=status && mode!=statusclear) {
            Connection->ReplyMessageKeyEvent("Message",k);
        }
    }
    return true;
}

bool cCommandProcessor::CmdNewMenu(cCommand &cmd) {
    if (!cmd.HasMaxArgument(1)) {
        Connection->ReplyErrorUnknownParameter(cmd.GetArg(1));
        return false;
    }
    if (!cmd.HasMinArgument(1)) {
        Connection->ReplyErrorMissingParameter();
        return false;
    }

    Contexts.Add(new cOsdServerMenu(cmd.GetArg(0)),cmd.GetAssignVar());
    return true;
}

bool cCommandProcessor::CmdMenuAddOsdItem(cCommand &cmd, cOsdServerMenu *menu) {
    if (!cmd.HasMaxArgument(1)) {
        Connection->ReplyErrorUnknownParameter(cmd.GetArg(1));
        return false;
    }
    if (!cmd.HasMinArgument(1)) {
        Connection->ReplyErrorMissingParameter();
        return false;
    }

    cOsdServerOsdItem *item=new cOsdServerOsdItem(cmd.GetArg(0));
    Contexts.Add(item,cmd.GetAssignVar());
    menu->Add(item);
    return true;
}

bool cCommandProcessor::CmdMenuShow(cCommand &cmd, cOsdServerMenu *menu) {
    if (cmd.IsAssignment()) {
        Connection->ReplyErrorAssign();
        return false;
    }
    if (!cmd.HasMaxArgument(0)) {
        Connection->ReplyErrorUnknownParameter(cmd.GetArg(0));
        return false;
    }

    menu->Show();
    return true;
}

bool cCommandProcessor::CmdMenuSendState(cCommand &cmd, cOsdServerMenu *menu) {
    if (cmd.IsAssignment()) {
        Connection->ReplyErrorAssign();
        return false;
    }
    if (!cmd.HasMaxArgument(1)) {
        Connection->ReplyErrorUnknownParameter(cmd.GetArg(1));
        return false;
    }
    if (!cmd.HasMinArgument(1)) {
        Connection->ReplyErrorMissingParameter();
        return false;
    }

    if (strcmp(cmd.GetArg(0),"osEnd")) {
        menu->SendOSState(osEnd);
    } else if (strcmp(cmd.GetArg(0),"osBack")) {
        menu->SendOSState(osBack);
    } else if (strcmp(cmd.GetArg(0),"osContinue")) {
        menu->SendOSState(osContinue);
    } else {
        Connection->ReplyErrorUnknownState(cmd.GetArg(0));
        return false;
    }
    return true;
}

bool cCommandProcessor::CmdMenuEnableEvent(cCommand &cmd, cOsdServerMenu *menu) {
    if (cmd.IsAssignment()) {
        Connection->ReplyErrorAssign();
        return false;
    }

    for (int i=0;cmd.GetArg(i);i++) {
        if (0==strcmp(cmd.GetArg(i),"keyOk")) {
            menu->EnableKeyEvent(kOk);
        } else if (0==strcmp(cmd.GetArg(i),"close")) {
            menu->EnableCloseEvent();
        } else {
            Connection->ReplyErrorUnknownEvent(cmd.GetArg(i));
        }
    }
    return true;
}

bool cCommandProcessor::CmdMenuSleepEvent(cCommand &cmd, int &timeout, cOsdServerMenu *menu) {
    if (cmd.IsAssignment()) {
        Connection->ReplyErrorAssign();
        return false;
    }

    SleepEventTimeout.tv_sec=0; // means infinite
    char *par=cmd.ArgScanFlagParam("timeout");
    if (par) {
        gettimeofday(&SleepEventTimeout,NULL);
        SleepEventTimeout.tv_sec+=atoi(par);
    } else {
        char *par=cmd.ArgScanFlagParam("timeoutms");
        if (par) {
            gettimeofday(&SleepEventTimeout,NULL);
            int t=atoi(par);
            SleepEventTimeout.tv_sec+=t/1000;
            SleepEventTimeout.tv_usec+=1000*(t%1000);
            if (SleepEventTimeout.tv_usec>=1000000) {
                SleepEventTimeout.tv_usec-=1000000;
                SleepEventTimeout.tv_sec++;
            }
        }
    }

    if (!cmd.HasMaxArgument(0)) {
        Connection->ReplyErrorUnknownParameter(cmd.GetArg(0));
        return false;
    }

    SleepEventObj=menu;

    if (HandleSleepEvent(timeout)) {
        return false;
    }
    State=stateSleepEvent;
    return true;
}

bool cCommandProcessor::CmdMenuGetCurrent(cCommand &cmd, cOsdServerMenu *menu) {
    if (cmd.IsAssignment()) {
        Connection->ReplyErrorAssign();
        return false;
    }
    if (!cmd.HasMaxArgument(0)) {
        Connection->ReplyErrorUnknownParameter(cmd.GetArg(0));
        return false;
    }

    int current;
    const char *name;
    menu->GetCurrent(current,name);
    Connection->ReplyMessageMenuCurrent(current,name);
    return true;
}

bool cCommandProcessor::CmdMenuSetCurrent(cCommand &cmd, cOsdServerMenu *menu) {
    if (cmd.IsAssignment()) {
        Connection->ReplyErrorAssign();
        return false;
    }
    if (!cmd.HasMaxArgument(1)) {
        Connection->ReplyErrorUnknownParameter(cmd.GetArg(1));
        return false;
    }
    if (!cmd.HasMinArgument(1)) {
        Connection->ReplyErrorMissingParameter();
        return false;
    }

    int NewCurrent=atoi(cmd.GetArg(0));

    menu->SetCurrent(NewCurrent);
    return true;
}


bool cCommandProcessor::CmdItemSetCurrent(cCommand &cmd, cOsdServerMenuItem *menuitem) {
    if (cmd.IsAssignment()) {
        Connection->ReplyErrorAssign();
        return false;
    }
    if (!cmd.HasMaxArgument(0)) {
        Connection->ReplyErrorUnknownParameter(cmd.GetArg(0));
        return false;
    }

    cOsdServerMenu *menu=menuitem->GetMenu();
    int NewCurrent=menuitem->GetIndex();

    menu->SetCurrent(NewCurrent);
    return true;
}



// Context based command dispatchers



bool cCommandProcessor::ProcessGlobalCommand(cCommand &cmd, int &timeout) {
    if (cmd.IsCommand("QUIT")) {
        if (CmdQuit(cmd)) {
            State=stateTerminated;
            return true;
        }
    } else if (cmd.IsCommand("IDLE")) {
        CmdIdle(cmd);
    } else if (cmd.IsCommand("MESSAGE")) {
        CmdMessage(cmd);
    } else if (cmd.IsCommand("NEWMENU")) {
        CmdNewMenu(cmd);
    } else {
        return false;
    }

    Connection->ReplyReady();
    return true;
}

bool cCommandProcessor::ProcessMenuCommand(cCommand &cmd, int &timeout, cOsdServerMenu *menu) {
    if (cmd.IsCommand("GETCURRENT")) {
        CmdMenuGetCurrent(cmd,menu);
    } else if (cmd.IsCommand("SETCURRENT")) {
        CmdMenuSetCurrent(cmd,menu);
    } else if (cmd.IsCommand("ADDOSDITEM")) {
        CmdMenuAddOsdItem(cmd,menu);
    } else if (cmd.IsCommand("SHOW")) {
        CmdMenuShow(cmd,menu);
    } else if (cmd.IsCommand("SENDSTATE")) {
        CmdMenuSendState(cmd,menu);
    } else if (cmd.IsCommand("ENABLEEVENT")) {
        CmdMenuEnableEvent(cmd,menu);
    } else if (cmd.IsCommand("SLEEPEVENT")) {
        if (CmdMenuSleepEvent(cmd,timeout,menu)) {
            return true;
        }
    } else {
        return false;
    }

    Connection->ReplyReady();
    return true;
}

bool cCommandProcessor::ProcessMenuItemCommand(cCommand &cmd, int &timeout, cOsdServerMenuItem *menuitem) {
    if (cmd.IsCommand("SETCURRENT")) {
        CmdItemSetCurrent(cmd,menuitem);
    } else {
        return false;
    }

    Connection->ReplyReady();
    return true;
}

bool cCommandProcessor::ProcessOsdItemCommand(cCommand &cmd, int &timeout, cOsdServerOsdItem *osditem) {
    return false;
}

bool cCommandProcessor::ProcessCommand(cCommand &cmd, int &timeout) {
    if (!cmd.GetContext()) {
        return ProcessGlobalCommand(cmd,timeout);
    }

    cOsdServerObject *ContextObj=Contexts.GetName(cmd.GetContext());
    if (!ContextObj) {
        Connection->ReplyErrorUnknownContext(cmd.GetContext());
        Connection->ReplyReady();
        return true;
    }

    switch (ContextObj->ClassId()) {
    case cOsdServerObject::clsidMenu:
        if (ProcessMenuCommand(cmd,timeout,ContextObj->Cast<cOsdServerMenu>())) return true;
        break;
    case cOsdServerObject::clsidOsdItem:
        if (ProcessMenuItemCommand(cmd,timeout,ContextObj->Cast<cOsdServerOsdItem>())) return true;
        if (ProcessOsdItemCommand(cmd,timeout,ContextObj->Cast<cOsdServerOsdItem>())) return true;
        break;
    }

    return false;
}


void cCommandProcessor::ProcessWelcome(cCommand &cmd, int &timeout) {
    if (cmd.IsCommand("QUIT")) {
        if (CmdQuit(cmd)) {
            State=stateTerminated;
            return;
        }
    } else if (cmd.IsCommand("VERSION")) {
        if (CmdVersion(cmd)) {
            State=stateReady;
            Connection->ReplyReady();
            return;
        }
    } else {
        Connection->ReplyErrorMissingVersion();
    }

    Connection->ReplyWelcome();
}

void cCommandProcessor::ProcessUnknown(cCommand &cmd) {
    if (cmd.GetCommand()) Connection->ReplyDebug("Parsed Command: \"%s\"",cmd.GetCommand());
    if (cmd.GetContext()) Connection->ReplyDebug("Parsed Context: \"%s\"",cmd.GetContext());
    if (cmd.GetAssignVar()) Connection->ReplyDebug("Parsed assignment to: \"%s\"",cmd.GetAssignVar());

    for (int i=0;cmd.GetArg(i);i++) {
        Connection->ReplyDebug("%s %i: \"%s\"",cmd.ArgIsFlag(i)?"Arg/Flag":"Argument",i,cmd.GetArg(i));
    }
    Connection->ReplyErrorUnknownCommand(cmd.GetCommand()?cmd.GetCommand():"");
    Connection->ReplyReady();
}



bool cCommandProcessor::HandleSleepEvent(int &timeout) {
    cOsdServerObject::cEvent event;

    if (SleepEventObj->PollEvent(event)) {
        if (event.ev==cOsdServerObject::cEvent::OkEvent) {
            Connection->ReplyMessageKeyEvent(event.who->Name(),kOk);
            return true;
        }
        if (event.ev==cOsdServerObject::cEvent::CloseEvent) {
            Connection->ReplyMessageEvent(event.who->Name(),"close");
            return true;
        }
    }

    if (SleepEventObj->IsDetached()) {
        Connection->ReplyErrorObjectNotActive();
        return true;
    }

    if (SleepEventTimeout.tv_sec==0) {
        timeout=100;
        return false;
    }

    timeval t;
    gettimeofday(&t,NULL);
    int ms=1000*(SleepEventTimeout.tv_sec-t.tv_sec)
           +(SleepEventTimeout.tv_usec-t.tv_usec)/1000;

    if (ms<=0) {
        Connection->ReplyMessageEventTimeout();
        return true;
    }

    timeout=min(ms,100);
    return false;
}


// Main command handler


bool cCommandProcessor::Action(cSelect &Selector) {

    for (;;) {
        int timeout=0;
        char *line;
        cCommand cmd;

        if (!Connection->IsConnected()) {
            State=stateTerminated;
        }

        switch (State) {
        case stateNew:
            Connection->ReplyWelcome();
            State=stateWelcome;
            break;

        case stateTerminated:
            return false;

        case stateSleepEvent:
            if (HandleSleepEvent(timeout)) {
                State=stateReady;
                Connection->ReplyReady();
            }
            break;

        case stateReady:
            line=Connection->ReadLine();
            if (!line) {
                Connection->AddSelect(Selector);
                return true;
            }
            if (cmd.ParseLine(*Connection,line)) {
                if (!ProcessCommand(cmd,timeout)) ProcessUnknown(cmd);
            }
            Connection->DelLine();
            break;

        case stateWelcome:
            line=Connection->ReadLine();
            if (!line) {
                Connection->AddSelect(Selector);
                return true;
            }
            if (cmd.ParseLine(*Connection,line)) {
                ProcessWelcome(cmd,timeout);
            }
            Connection->DelLine();
            break;
        } // end switch (State)

        if (timeout) {
            Selector.SetMinTimeoutMs(timeout);
            return true;
        }
    } // end for (;;)
    return false; // never reached
}

