#ifndef __OSDSERVER_INTERPRETER_H
#define __OSDSERVER_INTERPRETER_H

#include <vdr/tools.h>
#include <sys/time.h>
#include <time.h>

#include "tools.h"
#include "connection.h"
#include "osdobjects.h"



class cConnection : public cConnectionBase {
public:
    cConnection();

public:
    void ReplyReady() { Reply(200,"Ready."); }
    void ReplyWelcome() { Reply(201,"Welcome to OSDServer version %s, VDR version %s.",PLUGINVERSION,VDRVERSION); }
    void ReplyGoodBye() { Reply(202,"Good Bye."); }

    // error 400: Access Denied. See tools.c.
    void ReplyErrorIdentifier(const char *id) { Reply(401,"Error: Expected an identifier: %s",id); }
    void ReplyErrorIdentifierReserved(const char *id) { Reply(402,"Error: Reserved identifier: %s",id); }
    void ReplyErrorUnknownContext(const char *c) { Reply(403,"Error: Unknown context %s.",c); }
    void ReplyErrorUnknownCommand(const char *s) { Reply(404,"Error: Unknown command %s",s); }
    void ReplyErrorMissingVersion() { Reply(405,"Error: Need to specify VERSION first"); }
    void ReplyErrorUnknownVersion(const char *s) { Reply(406,"Error: Protocol version %s not supported.",s); }
    void ReplyErrorMissingParameter() { Reply(407,"Error: Missing Parameter."); }
    void ReplyErrorUnknownParameter(const char *par) { Reply(408,"Error: Unknown Parameter: %s",par); }
    void ReplyErrorAssign() { Reply(409,"Error: Command does not return an object to assign."); }
    void ReplyErrorParameter(const char *s) { Reply(410,"Error: Parameter Error: %s",s); }
    void ReplyErrorUnknownEvent(const char *s) { Reply(411,"Error: Unknown Event %s.",s); }
    void ReplyErrorObjectNotActive() { Reply(412,"Error: Object is not active."); }
    void ReplyErrorUnknownState(const char *s) { Reply(413,"Error: State is not known: %s.",s); }

    void ReplyMessageEvent(const char *who, const char *event) {
        Reply(300,"%s %s",who,event);
    }
    void ReplyMessageKeyEvent(const char *who, int k) {
        char buffer[32];
        ReplyMessageEvent(who,KeyString(buffer,sizeof buffer,k));
    }
    void ReplyMessageEventTimeout() {
        Reply(301,"timeout");
    }
    void ReplyMessageMenuCurrent(int i, const char *s) {
        if (s)
            Reply(302,"%i %s",i,s);
        else
            Reply(302,"%i",i);
    }

    static char* KeyString(char *buffer, int bufferlen, int key);
};



class cCommandProcessor : public cListObject {
public:
    cCommandProcessor(cConnection *Connection);
    ~cCommandProcessor();

    bool Action(cSelect &Selector);


    class cCommand {
    private:
        static const int MAXCOMMANDARGS=32;

        char *Command;
        char *Context;
        char *AssignVar;

        char *(args[MAXCOMMANDARGS]);
        int argc,argflagc;

    public:
        cCommand();

        bool ParseLine(cConnection &Connection, char *line);

        const char* GetCommand() { return Command; }
        const char* GetContext() { return Context; }
        const char* GetAssignVar() { return AssignVar; }

        bool IsCommand(const char *cmd) { return strcasecmp(Command,cmd)==0; }
        bool IsAssignment() { return AssignVar!=NULL; }
        bool HasMaxArgument(int max) { return argc<=max; }
        bool HasMinArgument(int min) { return argc>=min; }

        const char* GetArg(int i) { if (i>=0 && i<argc) return args[i]; else return NULL; }
        bool ArgIsFlag(int i) { return i>=0 && i<=argflagc && args[i][0]=='-'; }

        bool ArgScanFlag(const char *flag);
        char* ArgScanFlagParam(const char *flag);
    };

protected:
    cServerContext Contexts;

    cOsdServerMenu *SleepEventObj;
    timeval SleepEventTimeout;

    cConnection *Connection;

    enum enumState {
        stateNew,
        stateWelcome,
        stateReady,
        stateTerminated,
        stateSleepEvent
    };
    enumState State;


    bool CmdVersion(cCommand &cmd);
    bool CmdQuit(cCommand &cmd);
    bool CmdIdle(cCommand &cmd);


    bool CmdMessage(cCommand &cmd);
    bool CmdNewMenu(cCommand &cmd);

    bool CmdMenuAddOsdItem(cCommand &cmd, cOsdServerMenu *menu);
    bool CmdMenuShow(cCommand &cmd, cOsdServerMenu *menu);
    bool CmdMenuSendState(cCommand &cmd, cOsdServerMenu *menu);
    bool CmdMenuEnableEvent(cCommand &cmd, cOsdServerMenu *menu);
    bool CmdMenuSleepEvent(cCommand &cmd, int &timeout, cOsdServerMenu *menu);
    bool CmdMenuGetCurrent(cCommand &cmd, cOsdServerMenu *menu);
    bool CmdMenuSetCurrent(cCommand &cmd, cOsdServerMenu *menu);

    bool CmdItemSetCurrent(cCommand &cmd, cOsdServerMenuItem *menuitem);

    bool ProcessGlobalCommand(cCommand &cmd, int &timeout);
    bool ProcessMenuCommand(cCommand &cmd, int &timeout, cOsdServerMenu *menu);
    bool ProcessMenuItemCommand(cCommand &cmd, int &timeout, cOsdServerMenuItem *menuitem);
    bool ProcessOsdItemCommand(cCommand &cmd, int &timeout, cOsdServerOsdItem *osditem);
    void ProcessUnknown(cCommand &cmd);

    bool ProcessCommand(cCommand &cmd, int &timeout);
    void ProcessWelcome(cCommand &cmd, int &timeout);

    bool HandleSleepEvent(int &timeout);
};



#endif
